// Copyright (c) 2017, teja. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

library redstone.src.scanner;

import 'dart:mirrors';

import 'package:jaguar/jaguar.dart';
import 'package:jaguar_reflect/jaguar_reflect.dart';

part 'blacklist.dart';

class Ignore {
  const Ignore();
}

class _Scanner {
  final List<Symbol> filter;

  Set<Symbol> _libCache = new Set();

  _Scanner([this.filter]);

  Set<ClassMirror> scan() {
    final MirrorSystem mirrorSystem = currentMirrorSystem();
    final roots = new Set<LibraryMirror>();

    if (filter != null) {
      roots.addAll(filter.map(mirrorSystem.findLibrary));
    } else {
      roots.add(mirrorSystem.isolate.rootLibrary);
    }

    final libraries = new Set<LibraryMirror>();
    libraries.addAll(roots);
    roots.map(_scanLibrary).forEach(libraries.addAll);

    final groups = new Set<ClassMirror>();
    libraries.map(_loadHandlers).forEach(groups.addAll);

    return groups;
  }

  bool canSearchLibrary(LibraryDependencyMirror dep) {
    // Ignore exports
    if (!dep.isImport) return false;

    // Ignore black listed libraries
    return !blacklistSet.contains(dep.targetLibrary.simpleName);
  }

  Set<LibraryMirror> _scanLibrary(LibraryMirror mirror) {
    final ret = new Set<LibraryMirror>();

    for (final LibraryDependencyMirror dep in mirror.libraryDependencies) {
      if (!canSearchLibrary(dep)) continue;

      // If the library is already searched, dont search it again
      if (_libCache.contains(dep.targetLibrary.simpleName)) continue;
      _libCache.add(dep.targetLibrary.simpleName);

      if (dep.metadata.any(_isIgnored)) continue;

      ret.add(dep.targetLibrary);
      ret.addAll(_scanLibrary(dep.targetLibrary));
    }

    return ret;
  }

  bool _isIgnored(InstanceMirror im) => im.reflectee is Ignore;

  bool _isApi(InstanceMirror im) => im.reflectee is Api;

  Set<ClassMirror> _loadHandlers(LibraryMirror lib) {
    final apis = new Set<ClassMirror>();
    for (final DeclarationMirror declaration in lib.declarations.values) {
      if (declaration is! ClassMirror) continue;

      if (declaration.metadata.any(_isIgnored)) continue;

      if (declaration.metadata.any(_isApi)) {
        apis.add(declaration);
      }
    }
    return apis;
  }
}

void bootstrap(Jaguar server, {List<Symbol> filter}) {
  _Scanner scanner = new _Scanner(filter);

  Set<ClassMirror> apis = scanner.scan();

  bool _isConstructor(DeclarationMirror dm) =>
      dm is MethodMirror && dm.isConstructor && MirrorSystem.getName(dm.constructorName).isEmpty;

  MethodMirror _constructors(ClassMirror cm) =>
      cm.declarations.values.firstWhere(_isConstructor, orElse: () => null);

  for (final ClassMirror cm in apis) {
    DeclarationMirror constructor = cm.declarations[cm.simpleName];
    if(constructor is MethodMirror && constructor.parameters.length == 0) {
      final api = cm.newInstance(new Symbol(''), []).reflectee;
      server.addApi(reflectJaguar(api));
    } else {
      throw new Exception("Only Api with unnamed constructor with no parameters is supported at this time!");
    }
  }
}
